// MessageQueue.cpp: implementation of the MessageQueue class.
//
//////////////////////////////////////////////////////////////////////

#include "MessageQueue.h"
#include "nt_tcp_global_cpp.h"
#include <stdio.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

void DeleteHeapList(MessageQueue::InternalListNode *listnode)
{
	if (listnode == NULL)
		return;
	DeleteHeapList(listnode->next);
	// The queue didn't allocate the buffers so it doesn't free them either
	//delete listnode->buffer;
	delete listnode;
}

void DeleteHeap(MessageQueue::InternalNode *node)
{
	if (node == NULL)
		return;
	DeleteHeap(node->left);
	DeleteHeap(node->right);
	DeleteHeapList(node->list);
	delete node;
}

MessageQueue::MessageQueue()
{
	m_pHeap = NULL;
	m_hMutex = CreateMutex(NULL, FALSE, NULL);
	m_hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
}

MessageQueue::~MessageQueue()
{
	DeleteHeap(m_pHeap);
	m_pHeap = NULL;
	CloseHandle(m_hMutex);
	CloseHandle(m_hEvent);
}

bool InsertNode(MessageQueue::InternalNode *&node, int tag, void *buffer, int length, int from)
{
	if (node == NULL)
	{
		node = new MessageQueue::InternalNode;
		node->list = new MessageQueue::InternalListNode;
		node->list->buffer = buffer;
		node->list->length = length;
		node->list->from = from;
		node->list->next = NULL;
		node->left = NULL;
		node->right = NULL;
		node->tag = tag;
		return true;
	}

	if (tag == node->tag)
	{
		MessageQueue::InternalListNode *list = node->list;
		if (list == NULL)
		{
			node->list = new MessageQueue::InternalListNode;
			node->list->buffer = buffer;
			node->list->length = length;
			node->list->from = from;
			node->list->next = NULL;
			return true;
		}
		while (list->next != NULL)
			list = list->next;
		list->next = new MessageQueue::InternalListNode;
		list = list->next;
		list->next = NULL;
		list->buffer = buffer;
		list->length = length;
		list->from = from;
		return true;
	}

	if (tag < node->tag)
	{
		return InsertNode(node->left, tag, buffer, length, from);
	}

	if (tag > node->tag)
	{
		return InsertNode(node->right, tag, buffer, length, from);
	}

	return false;
}

bool MessageQueue::Insert(int tag, void *buffer, int length, int from)
{
	bool ret_val;
	if (WaitForSingleObject(m_hMutex, MESSAGE_QUEUE_TIMEOUT) == WAIT_TIMEOUT)
		return false;
	ret_val = InsertNode(m_pHeap, tag, buffer, length, from);
	// Every time data is inserted m_hEvent gets signalled
	if (ret_val)
	{
		DPRINTF(("MESSAGEQUEUE: tag %d, length %d inserted.\n", tag, length));
		SetEvent(m_hEvent);
	}
	ReleaseMutex(m_hMutex);
	return ret_val;
}

void* FindAndRemove(MessageQueue::InternalNode *node, int tag, int &length, int &from)
{
	if (node == NULL)
	{
		length = 0;
		return NULL;
	}
	
	if (tag == node->tag)
	{
		if (node->list == NULL)
			return NULL;
		MessageQueue::InternalListNode *list = node->list;
		void *buffer = node->list->buffer;
		length = node->list->length;
		from = node->list->from;
		node->list = node->list->next;
		delete list;
		return buffer;
	}

	if (tag < node->tag)
		return FindAndRemove(node->left, tag, length, from);

	if (tag > node->tag)
		return FindAndRemove(node->right, tag, length, from);

	from = 0;
	length = 0;
	return NULL;
}

void* MessageQueue::Remove(int tag, int &length, int &from)
{
	void *ret_val;
	if (WaitForSingleObject(m_hMutex, MESSAGE_QUEUE_TIMEOUT) == WAIT_TIMEOUT)
		return NULL;
	ret_val = FindAndRemove(m_pHeap, tag, length, from);
	// When a request for data fails m_hEvent is reset
	if (ret_val == NULL)
	{
		DPRINTF(("MESSAGEQUEUE: tag %d not available.\n", tag));
		ResetEvent(m_hEvent);
	}
	else
	{
		DPRINTF(("MESSAGEQUEUE: tag %d length %d removed.\n", tag, length));
	}
	ReleaseMutex(m_hMutex);
	return ret_val;
}

int FindAndReport(MessageQueue::InternalNode *node, int tag, int &from)
{
	if (node == NULL)
		return 0;

	if (tag == node->tag)
	{
		if (node->list == NULL)
			return 0;
		//return node->list->length;
		from = node->list->from;
		return 1;
	}

	if (tag < node->tag)
		return FindAndReport(node->left, tag, from);

	if (tag > node->tag)
		return FindAndReport(node->right, tag, from);

	return 0;
}

int MessageQueue::CountAvailable(int tag, int &from)
{
	int ret_val;
	if (WaitForSingleObject(m_hMutex, MESSAGE_QUEUE_TIMEOUT) == WAIT_TIMEOUT)
		return 0;
	ret_val = FindAndReport(m_pHeap, tag, from);
	ReleaseMutex(m_hMutex);
	return ret_val;
}

bool FindAvailable(MessageQueue::InternalNode *node, int tag, int &from)
{
	if (node == NULL)
		return false;

	if (tag == node->tag)
	{
		if (node->list == NULL)
			return false;
		from = node->list->from;
		return true;
	}

	if (tag < node->tag)
		return FindAvailable(node->left, tag, from);

	if (tag > node->tag)
		return FindAvailable(node->right, tag, from);

	return false;
}

bool MessageQueue::Available(int tag, int &from)
{
	bool ret_val;
	if (WaitForSingleObject(m_hMutex, MESSAGE_QUEUE_TIMEOUT) == WAIT_TIMEOUT)
		return false;
	ret_val = FindAvailable(m_pHeap, tag, from);
	ReleaseMutex(m_hMutex);
	return ret_val;
}

